<?php
namespace Zeedhi\Tools\EntityGenerator;

use Doctrine\Common\Annotations\Annotation;
use Doctrine\DBAL\Types\Type;
use Doctrine\ORM\Id\SequenceGenerator;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Doctrine\ORM\Mapping\MappingException;
use Doctrine\ORM\Tools\Export\Driver\AnnotationExporter;
use Doctrine\ORM\Tools\Export\Driver\XmlExporter;
use Doctrine\ORM\Tools\Export\Driver\YamlExporter;
use Zeedhi\Tools\Nominator;
use Zeedhi\Tools\Holder;

class MappingGenerator {

    const DRIVER_XML = 'xml';
    const DRIVER_YAML = 'yaml';
    const DRIVER_ANNOTATION = 'annotation';

    protected static $fieldZeedhiTypesToDoctrine = array(
        Holder\EntityField::TYPE_BLOB     => Type::BLOB,
        Holder\EntityField::TYPE_TEXT     => Type::TEXT,
        Holder\EntityField::TYPE_BOOLEAN  => Type::BOOLEAN,
        Holder\EntityField::TYPE_DATE     => Type::DATE,
        Holder\EntityField::TYPE_DATETIME => Type::DATETIME,
        Holder\EntityField::TYPE_FLOAT    => Type::FLOAT,
        Holder\EntityField::TYPE_INTEGER  => Type::INTEGER,
        Holder\EntityField::TYPE_OBJECT   => Type::OBJECT,
        Holder\EntityField::TYPE_STRING   => Type::STRING,
    );

    /** @var string */
    protected $driver;

    function __construct($verbose) {
        $this->driver = self::DRIVER_XML;
        $this->verbose = $verbose;
    }

    /**
     * @param string $driver
     */
    public function setDriver($driver) {
        $this->driver = $driver;
    }

    protected function log($message, array $context = array()) {
        if($this->verbose){
            echo "[log] ", str_replace(array_keys($context), $context, $message), PHP_EOL;
        }
    }

    protected function warning($message, array $context = array()) {
        echo "[warning] ", str_replace(array_keys($context), $context, $message), PHP_EOL;
    }

    /**
     * @param Holder\Entity $entity
     *
     * @return array Se doc of XMLExporter::setPrimaryTable
     */
    protected function getTableMapping(Holder\Entity $entity) {
        return array(
            "name" => $entity->getTable()->getName(),
        );
    }

    /**
     * @param ClassMetadata[] $metadata
     * @param string          $directory
     *
     * @throws \Exception If driver is invalid.
     *
     * @return void
     */
    protected function export($metadata, $directory) {
        switch($this->driver) {
            case self::DRIVER_YAML:
                $exporter = new YamlExporter($directory);
                break;
            case self::DRIVER_ANNOTATION:
                $exporter = new AnnotationExporter($directory);
                break;
            case self::DRIVER_XML:
                $exporter = new XmlExporter($directory);
                break;
            default:
                throw new \Exception("Invalid Driver {$this->driver}");
                break;
        }

        $exporter->setOverwriteExistingFiles(true);
        $exporter->setMetadata($metadata);
        $exporter->export();
    }

    /**
     * @param Holder\Entity $entity
     *
     * @return ClassMetadata
     */
    protected function createMetadataForEntity(Holder\Entity $entity) {
        $metadata = new ClassMetadata($entity->getClass()->getFullName());
        $metadata->setCustomRepositoryClass($entity->getRepositoryClass()->getFullName());
        $metadata->setPrimaryTable($this->getTableMapping($entity));

        $this->mapEntityFields($entity, $metadata);
        $this->mapEntityRelations($entity, $metadata);
        $this->log("Class metadata created for entity {entity}.", array("entity" => $entity->getName()));
        return $metadata;
    }

    /**
     * @param Holder\Entity[] $entitiesNames
     *
     * @return ClassMetadata[]
     */
    protected function createMetadataForEntities($entitiesToUse) {
        $metadata = array();
        foreach ($entitiesToUse as $entityName=>$entity) {
            try {
                if(isset($this->metaData["ENTITY"][$entityName])){
                    $metadata[] = $this->createMetadataForEntity($entity);
                }
            } catch (MappingException $e) {
                $this->warning("Entity skipped {entity} for reason {reason}.", array(
                    "entity" => $entityName,
                    "reason" => $e->getMessage(),
                ));
            }
        }

        return $metadata;
    }

    /**
     * Generates mapping (xml, yaml or php) for entities.
     *
     * @param Holder\Entity[] $entitiesNames    The entities names to generate the mapping.
     * @param string          $directory        The mapped directory.
     * @param mixed[]         $metaData         The entities, fields, entity fields and entitiy relations objects.
     * @param mixed[]         $data             The tables, constraints and sequences objects.
     *
     * @return boolean Return true if success.
     */
    public function generateMappingForEntities($entitiesToUse, $directory, $metaData, $data) {
        $this->log("Started to create mapping files.");
        $this->data = $data;
        $this->metaData = $metaData;
        $metadata = $this->createMetadataForEntities($entitiesToUse);
        $this->export($metadata, $directory);
        $this->log("{nrFiles} file(s) created.", array("nrFiles" => count($metadata)));
        return true;
    }

    /**
     * @param Holder\Entity $entity
     * @param $metadata
     */
    protected function mapEntityFields(Holder\Entity $entity, ClassMetadata $metadata) {
        $table = $entity->getTable();
        $tableName = $table->getName();
        $tableColumns = array_map(function($value){
            return $value->getName();
        }, $table->getColumns());

        /** @var Holder\EntityField[] $entityFields */
        $entityFields = $this->metaData["FIELD"][$tableName];

        $sequences = false;
        if(isset($this->data["SEQUENCES"]) && isset($this->data["SEQUENCES"][$tableName])){
            $sequences = $this->data["SEQUENCES"][$tableName];
        }

        foreach ($entityFields as $columnName => $entityField) {
            $isFk = $entityField->getIsFk() == 'Y';
            $isId = $entityField->getIsId() == 'Y';
            $column = $entityField->getColumn();

            if ($isFk && !$isId) continue;

            if ($isId) {
                if ($sequences) {
                    foreach ($sequences as $sequence) {
                        $metadata->setIdGeneratorType(ClassMetadataInfo::GENERATOR_TYPE_SEQUENCE);
                        $metadata->setIdGenerator(new SequenceGenerator($sequence->getName(), $sequence->getIncrementBy()));
                        $metadata->setSequenceGeneratorDefinition(array(
                            'sequenceName' => $sequence->getName(),
                            'allocationSize' => $sequence->getIncrementBy(),
                            'initialValue' => $sequence->getMinValue()
                        ));
                    }
                }
            }

            $mapping = array(
                "fieldName" => $entityField->getName(),
                "type" => self::$fieldZeedhiTypesToDoctrine[$entityField->getType()],
                "columnName" => $columnName,
                "id" => $isId,
                "associationKey" => $isId && $isFk
            );

            if($entityField->getIsUnique() == "Y"){
                $mapping["unique"] = true;
            }

            $metadata->mapField($mapping);

        }
    }

    protected function createJoinColumnMapping($entityRelation) {
        $targets = [];
        $relatedEntity = $entityRelation->getRelatedEntity();
        if($relatedEntity){
            $tableName = $relatedEntity->getTable()->getName();
            $constraint = $this->data["CONSTRAINTS"]["P"][$tableName][0];
            $targetConsColumns = $constraint->getConsColumns();
            foreach ($targetConsColumns as $targetConsColumn) {
                if($targetConsColumn){
                    $target = $targetConsColumn->getColumnName() ?: "ID";
                    $targets[] = [
                        "name" => $entityRelation->getEntityField()->getColumn()->getName(),
                        "referencedColumnName" => $target
                    ];
                }
            }
            return $targets;
        }
    }

    /**
     * @param Holder\Entity      $entity
     * @param ClassMetadata $metadata
     */
    protected function mapEntityRelations(Holder\Entity $entity, ClassMetadata $metadata) {
        /** @var Holder\EntityRelation[] $entityRelations */
        $entityName = $entity->getName();

        $table = $entity->getTable();
        $tableName = $table->getName();

        if(isset($this->metaData["ENTITYRELATION"][$tableName])){
            foreach ($this->metaData["ENTITYRELATION"][$tableName] as $constraintName => $columnsRelations) {
                foreach ($columnsRelations as $columnName => $entityRelation) {
                    $relatedEntity = $entityRelation->getRelatedEntity();
                    if($columnName){
                        $relationMapping = array(
                            "fieldName"     => $entityRelation->getName(),
                        );

                        $relationMapping["targetEntity"] = $relatedEntity
                            ? $relatedEntity->getClass()->getFullName()
                            : $entity->getClass()->getFullName();

                        if ($entityRelation->getIsInversed() === 'Y') {
                            $relationMapping["mappedBy"] = $entityRelation->getInverseRelation()->getEntityField()->getName();
                            $metadata->mapOneToMany($relationMapping);
                        } else {
                            $joinColumns = $this->createJoinColumnMapping($entityRelation);
                            if(!$joinColumns) continue;
                            foreach ($joinColumns as $joinColumn) {
                                $relationMapping["joinColumns"][] = $joinColumn;
                            }
                            $metadata->mapManyToOne($relationMapping);
                        }
                    }
                }
            }
        }
    }
}